/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

/* A clean and simple hash table implementation */

#include <unistd.h>
#include <stdio.h>
#include "symbolstable.h"

#define IDENTIFIER_SIZE 1000
#define TABLE_SIZE      1024

struct pair{
  char name[IDENTIFIER_SIZE+1];
  int value;
};

struct list{
  struct pair p;
  struct list *next;
};

struct symbols_table{
  struct list* lists[TABLE_SIZE];
  int next_position;
};

/* strnlen is a GNU extension, not all systems have it: */
size_t strnlen (const char *S, size_t MAXLEN){
  return ((strlen(S)<MAXLEN)?strlen(S):MAXLEN);
}

int hash_function(const char *s){
  int r=0;
  int i;

  for(i=0;i<strnlen(s,IDENTIFIER_SIZE);i++)
    r+=s[i]*i;

  return r % TABLE_SIZE;
}

int st_insert(struct symbols_table *s, const char *name, int value){
  int position=hash_function(name);
  struct list* element=s->lists[position];
  
  while(element!=NULL){
    if(!strncmp(element->p.name,name,IDENTIFIER_SIZE)){
      element->p.value=value;
      return value;
    }
    element=element->next;
  } /* while */

  /* If we arrived here then the element is not present. */

  element=(struct list*)malloc(sizeof(struct list));
    
  strncpy(element->p.name,name,IDENTIFIER_SIZE);
  element->p.value=value;
  element->next=s->lists[position];
  s->lists[position]=element;

  return value;
}

int st_insert_next(struct symbols_table *s, const char *name){
  return st_insert(s,name,s->next_position++);
}

int st_lookup(const struct symbols_table *s, const char *name){
  struct list* x=s->lists[hash_function(name)];

  while(x!=NULL){
    if(!strncmp(x->p.name,name,IDENTIFIER_SIZE))
      return x->p.value;
    x=x->next;
  }
  return -1;
}

void destroy_list(struct list *l){
  if(l->next!=NULL)
    destroy_list(l->next);
  free(l);
}

void st_destroy(struct symbols_table *s){
  int i;

  for(i=0;i<TABLE_SIZE;i++)
    if(s->lists[i]!=NULL)
      destroy_list(s->lists[i]);
  free(s);
}

struct symbols_table* st_create(){
  int i;
  struct symbols_table* s=(struct symbols_table*)
                          malloc(sizeof(struct symbols_table));

  for(i=0;i<TABLE_SIZE;i++)
    s->lists[i]=NULL;
  s->next_position=0;
}





